package asd.td7;

import asd.base.ObjetAClef;

/**
 * Une classe pour l'exercice 3 de la feuille 7.
 * Dans un tas d-aire :
 * - la racine du tas a pour indice 0
 * - le k-ieme fils du noeud d'indice i a pour indice d*i + k
 * - le pre du noeud d'indice j a pour indice  (j - 1) / d
 * Seules la mthode tasser et les constructeurs changent.
 */
public class TasDaireDecroissant {

    private ObjetAClef[] T; // le tableau implmentant le tas

    // le nombre d'lments prsents dans le Tas
    // et donc le numro du prochain noeud libre
    private int taille;

    // l'arit du tas
    private int d;

    /**
     * Construit un Tas d'arit d contenant les lments du tableau T.
     * Les mthodes de la classe Tas modifient les valeurs de T.
     * Le Tas occupe un "prfixe" de T, prcisment la "tranche"
     * T[0..i - 1] o i = taille(), le nombre d'lments prsents
     * dans le Tas. A l'instanciation le Tas occupe tout le tableau,
     * soit la tranche T[0..T.length - 1].
     * Complexit : O(n) o n = T.length.
     */
    public TasDaireDecroissant(ObjetAClef[] T, int d) {
        this.T = T;
        taille = T.length;
        this.d = d;
        for ( int n = pere(taille - 1); n >= 0; n-- )
            tasser(n);
    }

    /**
     * Construit un Tas vide de capacit maximum n.
     */
    public TasDaireDecroissant(int n, int d) {
        T = new ObjetAClef[n];
        taille = 0;
        this.d = d;
    }

    /**
     * Retourne le nombre d'lments prsents dans le Tas.
     */
    public int taille() {
        return taille;
    }

    /**
     * Retourne l'lment maximum du Tas.
     */
    public ObjetAClef maximum() {
	return T[0];
    }

    /**
     * Extrait et retourne l'lment maximum du Tas.
     * Prcondition : le Tas n'est pas vide.
     * Postrelation : si le Tas occupait la tranche [0..i] avant l'appel
     * de la mthode, aprs l'appel de la mthode le Tas occupe la tranche
     * [0..i - 1].
     * Complexit : O(lg(k)) o k  est le nombre d'lments dans le Tas.
     */
    public ObjetAClef extraireMaximum() {
        ObjetAClef max = T[0];
	T[0] = T[--taille]; // taille = taille - 1; T[0] = T[taille]
	if ( taille != 0 )
	    tasser(0);
        return max;
    }

    /**
     * Ajoute le nouvel lment x dans le Tas.
     * Prcondition : plein() == false.
     * Complexit : O(lg(k)) o k est le nombre d'lments dans le Tas.
     */
    public void ajouter(ObjetAClef x) {
        if ( taille < T.length ) {
            int n = taille++; // n = taille; taille = taille + 1;
            while ( n > 0 && T[pere(n)].clef() < x.clef() ) {
                T[n] = T[pere(n)];
                n = pere(n);
            }
            T[n] = x;
        }
    }

    /**
     * Donne la proprit de Tas au sous-arbre de noeud n.
     * Prcondition : N >= 0 et les sous-arbres de noeuds 2*n+1 et 2*n+2
     * sont des Tas.
     * Complexit : O(lg(k)) o k est le nombre d'lments contenus dans
     * l'arbre de noeud n.
     */
    private void tasser(int n) {
	int i = n;
        do {
            echanger(i,n); // ces deux lignes n'ont aucun effet lors du 
            n = i;         // premier passage dans la boucle
            for ( int k = d*n+1; k < taille && k <= d*(n+1); k++ )
                if ( T[k].clef() > T[i].clef() )
                    i = k;
        } while ( i != n );
    }

    /**
     * Retourne le numro du pre du noeud de numro n.
     */
    private int pere(int n) {
        return (n - 1) / d;
    }

    /**
     * Permute T[i] et T[j]
     */
    private void echanger(int i, int j) {
        ObjetAClef tmp = T[i];
        T[i] = T[j];
        T[j] = tmp;
    }
}
